@@ -7,8 +7,8 @@ use crate::elf;
 use crate::endian::{self, Endianness};
 use crate::pod::Pod;
 use crate::read::{
-    self, Error, ReadRef, Relocation, RelocationEncoding, RelocationKind, RelocationTarget,
-    SymbolIndex,
+    self, Error, ReadRef, Relocation, RelocationEncoding, RelocationFlags, RelocationKind,
+    RelocationTarget, SectionIndex, SymbolIndex,
 };
 
 use super::{ElfFile, FileHeader, SectionHeader, SectionTable};
@@ -23,10 +23,10 @@ impl RelocationSections {
     /// Create a new mapping using the section table.
     ///
     /// Skips relocation sections that do not use the given symbol table section.
-    pub fn parse<Elf: FileHeader>(
+    pub fn parse<'data, Elf: FileHeader, R: ReadRef<'data>>(
         endian: Elf::Endian,
-        sections: &SectionTable<Elf>,
-        symbol_section: usize,
+        sections: &SectionTable<'data, Elf, R>,
+        symbol_section: SectionIndex,
     ) -> read::Result<Self> {
         let mut relocations = vec![0; sections.len()];
         for (index, section) in sections.iter().enumerate().rev() {
@@ -34,7 +34,7 @@ impl RelocationSections {
             if sh_type == elf::SHT_REL || sh_type == elf::SHT_RELA {
                 // The symbol indices used in relocations must be for the symbol table
                 // we are expecting to use.
-                let sh_link = section.sh_link(endian) as usize;
+                let sh_link = SectionIndex(section.sh_link(endian) as usize);
                 if sh_link != symbol_section {
                     continue;
                 }
@@ -91,21 +91,21 @@ impl<'data, Elf: FileHeader> Iterator for ElfRelaIterator<'data, Elf> {
     }
 }
 
-/// An iterator over the dynamic relocations for an `ElfFile32`.
+/// An iterator for the dynamic relocations in an [`ElfFile32`](super::ElfFile32).
 pub type ElfDynamicRelocationIterator32<'data, 'file, Endian = Endianness, R = &'data [u8]> =
     ElfDynamicRelocationIterator<'data, 'file, elf::FileHeader32<Endian>, R>;
-/// An iterator over the dynamic relocations for an `ElfFile64`.
+/// An iterator for the dynamic relocations in an [`ElfFile64`](super::ElfFile64).
 pub type ElfDynamicRelocationIterator64<'data, 'file, Endian = Endianness, R = &'data [u8]> =
     ElfDynamicRelocationIterator<'data, 'file, elf::FileHeader64<Endian>, R>;
 
-/// An iterator over the dynamic relocations for an `ElfFile`.
+/// An iterator for the dynamic relocations in an [`ElfFile`].
 pub struct ElfDynamicRelocationIterator<'data, 'file, Elf, R = &'data [u8]>
 where
     Elf: FileHeader,
     R: ReadRef<'data>,
 {
     /// The current relocation section index.
-    pub(super) section_index: usize,
+    pub(super) section_index: SectionIndex,
     pub(super) file: &'file ElfFile<'data, Elf, R>,
     pub(super) relocations: Option<ElfRelaIterator<'data, Elf>>,
 }
@@ -130,9 +130,9 @@ where
             }
 
             let section = self.file.sections.section(self.section_index).ok()?;
-            self.section_index += 1;
+            self.section_index.0 += 1;
 
-            let sh_link = section.sh_link(endian) as usize;
+            let sh_link = SectionIndex(section.sh_link(endian) as usize);
             if sh_link != self.file.dynamic_symbols.section() {
                 continue;
             }
@@ -164,21 +164,21 @@ where
     }
 }
 
-/// An iterator over the relocations for an `ElfSection32`.
+/// An iterator for the relocations for an [`ElfSection32`](super::ElfSection32).
 pub type ElfSectionRelocationIterator32<'data, 'file, Endian = Endianness, R = &'data [u8]> =
     ElfSectionRelocationIterator<'data, 'file, elf::FileHeader32<Endian>, R>;
-/// An iterator over the relocations for an `ElfSection64`.
+/// An iterator for the relocations for an [`ElfSection64`](super::ElfSection64).
 pub type ElfSectionRelocationIterator64<'data, 'file, Endian = Endianness, R = &'data [u8]> =
     ElfSectionRelocationIterator<'data, 'file, elf::FileHeader64<Endian>, R>;
 
-/// An iterator over the relocations for an `ElfSection`.
+/// An iterator for the relocations for an [`ElfSection`](super::ElfSection).
 pub struct ElfSectionRelocationIterator<'data, 'file, Elf, R = &'data [u8]>
 where
     Elf: FileHeader,
     R: ReadRef<'data>,
 {
     /// The current pointer in the chain of relocation sections.
-    pub(super) section_index: usize,
+    pub(super) section_index: SectionIndex,
     pub(super) file: &'file ElfFile<'data, Elf, R>,
     pub(super) relocations: Option<ElfRelaIterator<'data, Elf>>,
 }
@@ -201,7 +201,7 @@ where
                 }
                 self.relocations = None;
             }
-            self.section_index = self.file.relocations.get(self.section_index)?;
+            self.section_index = SectionIndex(self.file.relocations.get(self.section_index.0)?);
             // The construction of RelocationSections ensures section_index is valid.
             let section = self.file.sections.section(self.section_index).unwrap();
             match section.sh_type(endian) {
@@ -239,35 +239,51 @@ fn parse_relocation<Elf: FileHeader>(
 ) -> Relocation {
     let mut encoding = RelocationEncoding::Generic;
     let is_mips64el = header.is_mips64el(endian);
+    let r_type = reloc.r_type(endian, is_mips64el);
+    let flags = RelocationFlags::Elf { r_type };
     let (kind, size) = match header.e_machine(endian) {
-        elf::EM_AARCH64 => match reloc.r_type(endian, false) {
-            elf::R_AARCH64_ABS64 => (RelocationKind::Absolute, 64),
-            elf::R_AARCH64_ABS32 => (RelocationKind::Absolute, 32),
-            elf::R_AARCH64_ABS16 => (RelocationKind::Absolute, 16),
-            elf::R_AARCH64_PREL64 => (RelocationKind::Relative, 64),
-            elf::R_AARCH64_PREL32 => (RelocationKind::Relative, 32),
-            elf::R_AARCH64_PREL16 => (RelocationKind::Relative, 16),
-            elf::R_AARCH64_CALL26 => {
-                encoding = RelocationEncoding::AArch64Call;
-                (RelocationKind::PltRelative, 26)
+        elf::EM_AARCH64 => {
+            if header.is_type_64() {
+                match r_type {
+                    elf::R_AARCH64_ABS64 => (RelocationKind::Absolute, 64),
+                    elf::R_AARCH64_ABS32 => (RelocationKind::Absolute, 32),
+                    elf::R_AARCH64_ABS16 => (RelocationKind::Absolute, 16),
+                    elf::R_AARCH64_PREL64 => (RelocationKind::Relative, 64),
+                    elf::R_AARCH64_PREL32 => (RelocationKind::Relative, 32),
+                    elf::R_AARCH64_PREL16 => (RelocationKind::Relative, 16),
+                    elf::R_AARCH64_CALL26 => {
+                        encoding = RelocationEncoding::AArch64Call;
+                        (RelocationKind::PltRelative, 26)
+                    }
+                    _ => (RelocationKind::Unknown, 0),
+                }
+            } else {
+                match r_type {
+                    elf::R_AARCH64_P32_ABS32 => (RelocationKind::Absolute, 32),
+                    _ => (RelocationKind::Unknown, 0),
+                }
             }
-            r_type => (RelocationKind::Elf(r_type), 0),
-        },
-        elf::EM_ARM => match reloc.r_type(endian, false) {
+        }
+        elf::EM_ARM => match r_type {
             elf::R_ARM_ABS32 => (RelocationKind::Absolute, 32),
-            r_type => (RelocationKind::Elf(r_type), 0),
+            _ => (RelocationKind::Unknown, 0),
         },
-        elf::EM_AVR => match reloc.r_type(endian, false) {
+        elf::EM_AVR => match r_type {
             elf::R_AVR_32 => (RelocationKind::Absolute, 32),
             elf::R_AVR_16 => (RelocationKind::Absolute, 16),
-            r_type => (RelocationKind::Elf(r_type), 0),
+            _ => (RelocationKind::Unknown, 0),
         },
-        elf::EM_BPF => match reloc.r_type(endian, false) {
+        elf::EM_BPF => match r_type {
             elf::R_BPF_64_64 => (RelocationKind::Absolute, 64),
             elf::R_BPF_64_32 => (RelocationKind::Absolute, 32),
-            r_type => (RelocationKind::Elf(r_type), 0),
+            _ => (RelocationKind::Unknown, 0),
         },
-        elf::EM_386 => match reloc.r_type(endian, false) {
+        elf::EM_CSKY => match r_type {
+            elf::R_CKCORE_ADDR32 => (RelocationKind::Absolute, 32),
+            elf::R_CKCORE_PCREL32 => (RelocationKind::Relative, 32),
+            _ => (RelocationKind::Unknown, 0),
+        },
+        elf::EM_386 => match r_type {
             elf::R_386_32 => (RelocationKind::Absolute, 32),
             elf::R_386_PC32 => (RelocationKind::Relative, 32),
             elf::R_386_GOT32 => (RelocationKind::Got, 32),
@@ -278,9 +294,9 @@ fn parse_relocation<Elf: FileHeader>(
             elf::R_386_PC16 => (RelocationKind::Relative, 16),
             elf::R_386_8 => (RelocationKind::Absolute, 8),
             elf::R_386_PC8 => (RelocationKind::Relative, 8),
-            r_type => (RelocationKind::Elf(r_type), 0),
+            _ => (RelocationKind::Unknown, 0),
         },
-        elf::EM_X86_64 => match reloc.r_type(endian, false) {
+        elf::EM_X86_64 => match r_type {
             elf::R_X86_64_64 => (RelocationKind::Absolute, 64),
             elf::R_X86_64_PC32 => (RelocationKind::Relative, 32),
             elf::R_X86_64_GOT32 => (RelocationKind::Got, 32),
@@ -295,38 +311,57 @@ fn parse_relocation<Elf: FileHeader>(
             elf::R_X86_64_PC16 => (RelocationKind::Relative, 16),
             elf::R_X86_64_8 => (RelocationKind::Absolute, 8),
             elf::R_X86_64_PC8 => (RelocationKind::Relative, 8),
-            r_type => (RelocationKind::Elf(r_type), 0),
+            _ => (RelocationKind::Unknown, 0),
         },
-        elf::EM_HEXAGON => match reloc.r_type(endian, false) {
+        elf::EM_HEXAGON => match r_type {
             elf::R_HEX_32 => (RelocationKind::Absolute, 32),
-            r_type => (RelocationKind::Elf(r_type), 0),
+            _ => (RelocationKind::Unknown, 0),
         },
-        elf::EM_MIPS => match reloc.r_type(endian, is_mips64el) {
+        elf::EM_LOONGARCH => match r_type {
+            elf::R_LARCH_32 => (RelocationKind::Absolute, 32),
+            elf::R_LARCH_64 => (RelocationKind::Absolute, 64),
+            elf::R_LARCH_32_PCREL => (RelocationKind::Relative, 32),
+            elf::R_LARCH_64_PCREL => (RelocationKind::Relative, 64),
+            elf::R_LARCH_B16 => {
+                encoding = RelocationEncoding::LoongArchBranch;
+                (RelocationKind::Relative, 16)
+            }
+            elf::R_LARCH_B21 => {
+                encoding = RelocationEncoding::LoongArchBranch;
+                (RelocationKind::Relative, 21)
+            }
+            elf::R_LARCH_B26 => {
+                encoding = RelocationEncoding::LoongArchBranch;
+                (RelocationKind::Relative, 26)
+            }
+            _ => (RelocationKind::Unknown, 0),
+        },
+        elf::EM_MIPS => match r_type {
             elf::R_MIPS_16 => (RelocationKind::Absolute, 16),
             elf::R_MIPS_32 => (RelocationKind::Absolute, 32),
             elf::R_MIPS_64 => (RelocationKind::Absolute, 64),
-            r_type => (RelocationKind::Elf(r_type), 0),
+            _ => (RelocationKind::Unknown, 0),
         },
-        elf::EM_MSP430 => match reloc.r_type(endian, false) {
+        elf::EM_MSP430 => match r_type {
             elf::R_MSP430_32 => (RelocationKind::Absolute, 32),
             elf::R_MSP430_16_BYTE => (RelocationKind::Absolute, 16),
-            r_type => (RelocationKind::Elf(r_type), 0),
+            _ => (RelocationKind::Unknown, 0),
         },
-        elf::EM_PPC => match reloc.r_type(endian, false) {
+        elf::EM_PPC => match r_type {
             elf::R_PPC_ADDR32 => (RelocationKind::Absolute, 32),
-            r_type => (RelocationKind::Elf(r_type), 0),
+            _ => (RelocationKind::Unknown, 0),
         },
-        elf::EM_PPC64 => match reloc.r_type(endian, false) {
+        elf::EM_PPC64 => match r_type {
             elf::R_PPC64_ADDR32 => (RelocationKind::Absolute, 32),
             elf::R_PPC64_ADDR64 => (RelocationKind::Absolute, 64),
-            r_type => (RelocationKind::Elf(r_type), 0),
+            _ => (RelocationKind::Unknown, 0),
         },
-        elf::EM_RISCV => match reloc.r_type(endian, false) {
+        elf::EM_RISCV => match r_type {
             elf::R_RISCV_32 => (RelocationKind::Absolute, 32),
             elf::R_RISCV_64 => (RelocationKind::Absolute, 64),
-            r_type => (RelocationKind::Elf(r_type), 0),
+            _ => (RelocationKind::Unknown, 0),
         },
-        elf::EM_S390 => match reloc.r_type(endian, false) {
+        elf::EM_S390 => match r_type {
             elf::R_390_8 => (RelocationKind::Absolute, 8),
             elf::R_390_16 => (RelocationKind::Absolute, 16),
             elf::R_390_32 => (RelocationKind::Absolute, 32),
@@ -365,16 +400,75 @@ fn parse_relocation<Elf: FileHeader>(
                 encoding = RelocationEncoding::S390xDbl;
                 (RelocationKind::GotBaseRelative, 32)
             }
-            r_type => (RelocationKind::Elf(r_type), 0),
+            _ => (RelocationKind::Unknown, 0),
         },
-        elf::EM_SPARC | elf::EM_SPARC32PLUS | elf::EM_SPARCV9 => {
-            match reloc.r_type(endian, false) {
-                elf::R_SPARC_32 | elf::R_SPARC_UA32 => (RelocationKind::Absolute, 32),
-                elf::R_SPARC_64 | elf::R_SPARC_UA64 => (RelocationKind::Absolute, 64),
-                r_type => (RelocationKind::Elf(r_type), 0),
+        elf::EM_SBF => match r_type {
+            elf::R_SBF_64_64 => (RelocationKind::Absolute, 64),
+            elf::R_SBF_64_32 => (RelocationKind::Absolute, 32),
+            _ => (RelocationKind::Unknown, 0),
+        },
+        elf::EM_SHARC => match r_type {
+            elf::R_SHARC_ADDR24_V3 => {
+                encoding = RelocationEncoding::SharcTypeA;
+                (RelocationKind::Absolute, 24)
             }
-        }
-        _ => (RelocationKind::Elf(reloc.r_type(endian, false)), 0),
+            elf::R_SHARC_ADDR32_V3 => {
+                encoding = RelocationEncoding::SharcTypeA;
+                (RelocationKind::Absolute, 32)
+            }
+            elf::R_SHARC_ADDR_VAR_V3 => {
+                encoding = RelocationEncoding::Generic;
+                (RelocationKind::Absolute, 32)
+            }
+            elf::R_SHARC_PCRSHORT_V3 => {
+                encoding = RelocationEncoding::SharcTypeA;
+                (RelocationKind::Relative, 6)
+            }
+            elf::R_SHARC_PCRLONG_V3 => {
+                encoding = RelocationEncoding::SharcTypeA;
+                (RelocationKind::Relative, 24)
+            }
+            elf::R_SHARC_DATA6_V3 => {
+                encoding = RelocationEncoding::SharcTypeA;
+                (RelocationKind::Absolute, 6)
+            }
+            elf::R_SHARC_DATA16_V3 => {
+                encoding = RelocationEncoding::SharcTypeA;
+                (RelocationKind::Absolute, 16)
+            }
+            elf::R_SHARC_DATA6_VISA_V3 => {
+                encoding = RelocationEncoding::SharcTypeB;
+                (RelocationKind::Absolute, 6)
+            }
+            elf::R_SHARC_DATA7_VISA_V3 => {
+                encoding = RelocationEncoding::SharcTypeB;
+                (RelocationKind::Absolute, 7)
+            }
+            elf::R_SHARC_DATA16_VISA_V3 => {
+                encoding = RelocationEncoding::SharcTypeB;
+                (RelocationKind::Absolute, 16)
+            }
+            elf::R_SHARC_PCR6_VISA_V3 => {
+                encoding = RelocationEncoding::SharcTypeB;
+                (RelocationKind::Relative, 16)
+            }
+            elf::R_SHARC_ADDR_VAR16_V3 => {
+                encoding = RelocationEncoding::Generic;
+                (RelocationKind::Absolute, 16)
+            }
+            _ => (RelocationKind::Unknown, 0),
+        },
+        elf::EM_SPARC | elf::EM_SPARC32PLUS | elf::EM_SPARCV9 => match r_type {
+            elf::R_SPARC_32 | elf::R_SPARC_UA32 => (RelocationKind::Absolute, 32),
+            elf::R_SPARC_64 | elf::R_SPARC_UA64 => (RelocationKind::Absolute, 64),
+            _ => (RelocationKind::Unknown, 0),
+        },
+        elf::EM_XTENSA => match r_type {
+            elf::R_XTENSA_32 => (RelocationKind::Absolute, 32),
+            elf::R_XTENSA_32_PCREL => (RelocationKind::Relative, 32),
+            _ => (RelocationKind::Unknown, 0),
+        },
+        _ => (RelocationKind::Unknown, 0),
     };
     let sym = reloc.r_sym(endian, is_mips64el) as usize;
     let target = if sym == 0 {
@@ -389,10 +483,11 @@ fn parse_relocation<Elf: FileHeader>(
         target,
         addend: reloc.r_addend(endian).into(),
         implicit_addend,
+        flags,
     }
 }
 
-/// A trait for generic access to `Rel32` and `Rel64`.
+/// A trait for generic access to [`elf::Rel32`] and [`elf::Rel64`].
 #[allow(missing_docs)]
 pub trait Rel: Debug + Pod + Clone {
     type Word: Into<u64>;
@@ -457,7 +552,7 @@ impl<Endian: endian::Endian> Rel for elf::Rel64<Endian> {
     }
 }
 
-/// A trait for generic access to `Rela32` and `Rela64`.
+/// A trait for generic access to [`elf::Rela32`] and [`elf::Rela64`].
 #[allow(missing_docs)]
 pub trait Rela: Debug + Pod + Clone {
     type Word: Into<u64>;
